module Polarity where

import Util
import Warshall

import Data.Map (Map)
import qualified Data.Map as Map
import qualified Data.List as List

-- polarities and strict positivity ----------------------------------

type PVarId = Int

data Pos 
  = SPos  -- strictly positive 
  | Pos   -- positive
  | Mixed -- unknown polarity (invariant)
  | Neg   -- negative, used internally for contravariance of sized codata  
  | NOcc  -- non-occurring, used to determine whether data type is recursive
  | PVar PVarId -- flexible polarity variable
         deriving (Eq,Ord)

instance Show Pos where
  show SPos  = "++"
  show Pos   = "+"
  show Mixed = "o"
  show Neg   = "-"
  show NOcc  = "/"
  show (PVar i) = showPVar i

showPVar i = "?p" ++ show i

isPVar (PVar{}) = True
isPVar _ = False

-- information ordering
leqPol :: Pos -> Pos -> Bool
leqPol x NOcc   = True   -- NOcc is top
leqPol NOcc x   = False
leqPol Mixed y  = True   -- Mixed is bottom
leqPol x Mixed  = False
leqPol Pos SPos = True 
leqPol x   y    = x == y

{- RETIRED
isSPos :: Pos -> Bool
isSPos SPos = True
isSPos NOcc = True
isSPos _ = False
-}

{- NOT USED
isPos :: Pos -> Bool
isPos Pos = True
isPos x = isSPos x
-}

-- polarity negation
-- used in Eval.hs leqVals' for switching sides
-- this means it is only applied to Pos,Neg,Mixed, 
-- never to SPos,NOcc, or polarity expressions
polNeg :: Pos -> Pos
polNeg SPos  = Neg
polNeg Pos   = Neg
polNeg Neg   = Pos
polNeg Mixed = Mixed
polNeg NOcc  = NOcc

-- polarity composition
-- used in Eval.hs leqVals'
polComp :: Pos -> Pos -> Pos
polComp NOcc  x   = NOcc   -- most dominant
polComp x NOcc    = NOcc
polComp Mixed x   = Mixed  -- dominant except for NOcc
polComp x Mixed   = Mixed
polComp SPos  x   = x      -- neutral
polComp x SPos    = x
polComp Pos  x    = x      -- neutral except for SPos
polComp x Pos     = x
polComp Neg Neg   = Pos    -- order 2
{- pol.comp. is ass., comm., with neutral ++, and infinity NOcc
   cancellation does not hold, since composition with anything by ++ is
   information loss:
     q p <= q p' ==> p <= p'
   only if q = ++ (then it is trivial anyway) -}

-- polarity inverse composition (see Abel, MSCS 2008)
-- invComp p q1 <= q2  <==> q1 <= polComp p q2 
-- used in TCM.hs cxtApplyDec
invComp :: Pos -> Pos -> Pos
invComp Mixed Mixed = Mixed
invComp Mixed x     = NOcc
invComp NOcc  x     = Mixed
invComp SPos  x     = x         -- SPos is the identity
invComp p     SPos  = NOcc      -- SPos preserved only under SPos
invComp Pos   x     = x         -- x not SPos
invComp Neg   x     = polNeg x  -- x not SPos

{- UNUSED
invCompExpr :: Pos -> PExpr -> PExpr
invCompExpr q (PValue p)   = PValue $ invComp q p
invCompExpr q (PExpr q' i) = PExpr (polComp q q') i 
-}

-- polarity conjuction (infimum)
-- currently unused
polAnd :: Pos -> Pos -> Pos
polAnd NOcc x       = x       -- most information
polAnd x NOcc       = x
polAnd x y | x == y = x       -- same information
polAnd SPos Pos     = Pos     -- SPos is more informative than Pos
polAnd Pos SPos     = Pos
polAnd _ _          = Mixed   -- remaining cases: conflicting or no info
{-
polAnd Mixed x  = Mixed   -- least information
polAnd x Mixed  = Mixed
polAnd SPos Neg = Mixed
polAnd Neg SPos = Mixed
-}

instance SemiRing Pos where
  oplus  = polAnd
  otimes = polComp
  ozero  = NOcc
  oone   = SPos  

relPol :: Pos -> (a -> a -> Bool) -> (a -> a -> Bool)
relPol NOcc  r a b = True
relPol Mixed r a b = r a b && r b a
relPol Neg   r a b = r b a
relPol Pos   r a b = r a b
relPol SPos  r a b = r a b

relPolM :: (Monad m) => Pos -> (a -> a -> m ()) -> (a -> a -> m ())
relPolM NOcc  r a b = return ()
relPolM Mixed r a b = r a b >> r b a
relPolM Neg   r a b = r b a
relPolM Pos   r a b = r a b
relPolM SPos  r a b = r a b

-- polarity product (composition of polarities) ----------------------

data Multiplicity = POne | PTwo deriving (Eq, Ord)

instance Show Multiplicity where
  show POne = "1"
  show PTwo = "2"

-- addition modulo 2
addMultiplicity :: Multiplicity -> Multiplicity -> Multiplicity
addMultiplicity PTwo y = y
addMultiplicity x PTwo = x
addMultiplicity POne POne = PTwo

type VarMults = Map PVarId Multiplicity -- multiplicity of variables (1 or 2) 

showMults :: VarMults -> String
showMults mults = 
  let ml = Map.toList mults  -- get list of (key,value) pairs
      l  = concat $ map f ml where
             f (k, POne) = [k]
             f (k, PTwo) = [k,k]
  in Util.showList "." showPVar l    

multsEmpty = Map.empty

multsSingle :: Int -> VarMults
multsSingle i = Map.insert i POne multsEmpty


data PProd = PProd 
  { coeff    :: Pos      -- a coefficient, excluding PVar
  , varMults :: VarMults -- multiplicity of variables (1 or 2) 
  } deriving (Eq,Ord)

instance Show PProd where
  show (PProd NOcc _) = "/"
  show (PProd SPos m) = if Map.null m then show SPos else showMults m
  show (PProd q m) = separate "." (show q) (showMults m)

pprod :: Pos -> PProd
pprod (PVar i) = PProd SPos (multsSingle i)
pprod q = PProd q multsEmpty

isSPos :: PProd -> Bool
isSPos (PProd NOcc _) = True
isSPos (PProd SPos m) = Map.null m
isSPos _ = False

-- multiply two products

polProd :: PProd -> PProd -> PProd
polProd (PProd q1 m1) (PProd q2 m2) = PProd (polComp q1 q2) $
  Map.unionWith addMultiplicity m1 m2

-- polarity expressions are polynomials ------------------------------

data PPoly = PPoly { monomials :: [PProd] } deriving (Eq,Ord)

instance Show PPoly where
  show (PPoly []) = show NOcc
  show (PPoly [m]) = show m
  show (PPoly l)   = Util.showList "/\\" show l

ppoly :: PProd -> PPoly
ppoly (PProd NOcc _) = PPoly []
ppoly pp = PPoly [pp]

polSum :: PPoly -> PPoly -> PPoly
polSum (PPoly x) (PPoly y) = PPoly $ List.nub $ x ++ y

polProduct :: PPoly -> PPoly -> PPoly
polProduct (PPoly l1) (PPoly l2) =
  let ps = [ polProd x y | x <- l1, y <- l2]
  in PPoly $ List.nub $ ps
 
instance SemiRing PPoly where
  oplus  = polSum
  otimes = polProduct
  ozero  = PPoly []
  oone   = PPoly [PProd SPos Map.empty]

{-
data PExpr 
  = PValue Pos     -- constant polarity
  | PExpr Pos Int  -- PExpr q pi means q^_1 pi  (pi is the number of the var)

-- a polarity variable
pvar :: Int -> PExpr
pvar = PExpr SPos  -- ++ is the neutral element of inverse polarity composition
  
instance Show PExpr where
  show (PValue p) = show p
  show (PExpr SPos i) = "?p" ++ show i
  show (PExpr q i) = show q ++ "^-1(?p" ++ show i ++ ")"
-}


{- ML-style Polarity inference

Preliminaries:
1. constructor types are mixed-variant function types only
2. matching is only allowed on mixed-variant arguments
  1+2 are both consequences that only type-valued functions have variance
  and 1. data constructors are not types, 2. types are not matched on

Concrete syntax

  f : (xs : As) -> C   (C not a Pi-type)
  f = t

is parsed as abstract syntax

  f : pis(xs : As) -> C
  f = t

where pi_1..n are fresh polarity variables

Then t is type-checked to infer the polarity variables, e.g.

  f xs = t

  pis(xs : As) |- t : C

Now what can happen?

Variable:  t = x_i.  Then we add a constraint  pi_i <= ++

Application t = u v  where u : q(x:B) -> D

  q^-1(pis(xs: As)) |- v : B

  A term q^-1 pi arises where q is a polarity constant (!, ML-inference)
  or a polarity variable (recursion!, e.g. u = f)
  and pi is a polarity expression

In the context, keep SOLL and HABEN

  SOLL  is the original polarity (variable or constant)
  HABEN is a (ordered) list of pol.vars. and a pol.const. (default: ++)

Variable   : add constraint SOLL <= HABEN
Application: add q to HABEN by polarity multiplication (q is a var or const)
Abstraction: \xt : q(x:A) -> B:  continue with x (SOLL = q, HABEN = ++)

What kind of constraints do arise
1) q  <= pi    [ from variables , pi is a Pol-product ]
2) ++ <= pis  [ from positivity graph, pis is a sum of Pol-products ]
   this means ++ <= pi for all pi in pis

Solving constraints
 
- discard  o <= pi  and q <= /  (do not even need to add them)
- all pvars which are not bounded below (appearing in one q in 1)
  can be instantiated to /  which will remove some constraints


-}

{- Mutual recursion 

In mutual declarations, use the following Ansatz:  data/codata ++, functions o

  A = B -> A
  B = A -> B

A (B) is positive in its own body and negative in the body of B (A)

  F A B = B -> A   F(-,++)
  G A B = A -> B   G(-,++)

  F A B = G A B -> F A B  
  G A B = F A B -> G A B

  Polarities: 
  F : fa * -> fb * -> *
  G : ga * -> gb * -> *  

  A : -fa, B : -fb |- G A B : *  ==> -fa <= ga, -fb <= gb
  A : -ga, B : -gb |- F A B : *  ==> -ga <= fa, -gb <= fb

-}

{- Pure polarity inference

Judgement:  pis(xs:As) |- t : B ---> C

Variable:   pis(xs:As) |- xi : Ai ---> pi_i <= ++

Application: Delta |- u : q(x:A) -> B ---> C1
             Delta |- v : A           ---> C2
             --------------------------------------------------
             Delta |- u v : B[u/x] ---> C1,C2,q(Delta) <= Delta
-}